/***********************************************************************
**  Function:   yi(iseed,m,idist,parm,n,rho,rhox)                     **
**                                                                    **
**  Purpose:    Estimate the correlation matrix of the random vector  **
**              X that is generated by the marginal oriented approach,**
**              using rho as the step-1 cholelation matrix.           **
**                                                                    **
**  Author:     Huifen Chen                                           **
**              Department of Industrial Engineering                  **
**              Da-Yeh Institute of Technology                        **
**              Chang-Hwa, 51505, TAIWAN                              **
**  Date:       December 30, 1994.                                    **
**              Updated from Chen and Schmeiser's code.               **
**                                                                    **
**  Input:                                                            **
**              iseed:   Random-number seed                           **
**              m:       Sample size of correlation estimate          **
**              parm:    Vector with distribution properties of ghat  **
**              n:       Dimension of the random vector X             **
**              rho:     Step-1 correlation matrix                    **
**                                                                    **
**  Output:                                                           **
**              rhox:    Estimated Step-3 correlation matrix          **
**                                                                    **
**  Called by:                                                        **
**              retro:   A stochastic root-finding algorithm          **
**                                                                    **
**  Routine called:                                                   **
**              multvar: Generation of multivariate vectors using     **
**                       the marginal oriendted approach, given the   **
**                       Step-1 correlation matrix.                   **
***********************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
/*
#include <malloc.h>
*/


void  yi(long *iseed,int m,int *idist,double **parm,int n,double **rho,
         double **rhox);
void corr_all(long *iseed,int m,int *idist,double **parm,int n,
              double **rho,double **chol,double **rhox);
void corr_pair(long *iseed,int m,int *idist,double **parm,int n,
               double **rho,double **rhox,int *ifault);
void rowcorr(int n,long *iseed,int m,int *idist,double **parm,
             double **rho,double **rhox);
void  multvar(long *iseed,int *idist,double **parm,int n,double **chol,
              double *z,double *x);
void  multnormal(long *iseed,int n,double **chol,double *z);

/* void  finv(); */

void ndtr(double x,double *p,double *d);
void open_array(int n,double **x,double **z,double **zbar,double **xbar,
                double ***rhoz,double **sumx,double **sumz,double ***sumx12,
                double ***sumz12,double ***sum1,double ***sum2,double ***sum3,
                double ***sum4,double **ssumx,double **ssumz,double ***ssumx12,
                double ***ssumz12);
void free_array(double *x,double *z,double *xbar,double *zbar,double *sumx,
                double *sumz,double *ssumx,double *ssumz,double **rhoz,
                double **sumx12,double **sumz12,double **ssumx12,
                double **ssumz12,double **sum1,double **sum2,double **sum3,
                double **sum4);
double u16807d(long *iseed);
double ppnd(double p,int *ifault );
void free_2darray(int m, int n, double **a);

void  yi(long *iseed,int m,int *idist,double **parm,int n,double **rho,
         double **rhox)
{
   int     i,ifault;
   double   **chol;

   /*  open local arrays  */
   chol   = (double**)malloc(n*sizeof(*chol));
   *chol  = (double *)malloc((n*n)*sizeof(**chol));
   for(i = 1;i < n;i++)
       *(chol+i) = *chol + i*n;

   /* ...compute cholesky decomposition of rho, so that a=chol'*chol,
   **    where chol is an upper triangular matrix.   */
   choleskyf(n,rho,chol,&ifault);

   if (ifault == 0)
       corr_all(iseed,m,idist,parm,n,rho,chol,rhox);

   else {

       printf("Cholesky decomposition has error indicator %d\n",ifault);
       corr_pair(iseed,m,idist,parm,n,rho,rhox,&ifault);
   } 

   free( *chol );
        free( chol );
}

   
void corr_all(long *iseed,int m,int *idist,double **parm,int n,
              double **rho,double **chol,double **rhox)
{

   int     i,j,k,l,batchsize;
   double   tmp,cov,beta,
           *x,*z,*zbar,*xbar,**rhoz;
   double  *sumx,*sumz,**sumx12,**sumz12,**sum1,**sum2,**sum3,**sum4,
           *ssumx,*ssumz,**ssumx12,**ssumz12;

   /*  open local arrays  */
   open_array(n,&x,&z,&zbar,&xbar,&rhoz,&sumx,&sumz,
              &sumx12,&sumz12,&sum1,&sum2,&sum3,&sum4,&ssumx,
              &ssumz,&ssumx12,&ssumz12);

   /*  initialize macro sums  */
   for(i=0;i < n;i++)
   {
       ssumz[i] = 0.;
       ssumx[i] = 0.;
       for(j=i;j < n;j++)
       {
           ssumz12[i][j]  = 0.;
           ssumx12[i][j]  = 0.;
           sum1[i][j]     = 0.;
           sum2[i][j]     = 0.;
           sum3[i][j]     = 0.;
           sum4[i][j]     = 0.;
           rhoz[i][j]     = 0.;
           rhox[i][j]     = 0.;
       }
   }



   batchsize = m / 20;      /* set macro replication number = 20 */
   /* ...if batchsize < 2, use the crude estimate of correlations */   
   /* ...generate m bivariate observations and estimate corr(X1,X2). */
  if (batchsize < 2)               
  {             
     for(i=0;i < n; i++)    /* initialize the zero micro-sum */  
     {  
        sumx[i]   = 0.;           
        for(j=i;j < n;j++)        
        {                         
            sumx12[i][j] = 0.;    
        }                         
     }                            
                                  
     for(k=0;k < m;k++)           
     {                            
         multvar(iseed,idist,parm,n,chol,z,x);  
         for(i=0;i < n;i++)       
         {                        
             sumx[i] = sumx[i] + x[i]; 
             for(j=i;j < n;j++)    
                 sumx12[i][j] = sumx12[i][j] + x[i] * x[j]; 
         }                        
     }                            
                                  
     for(i=0;i < n;i++)           
        xbar[i] = sumx[i] / (double) m;   
                                  
     for(i=0;i < n;i++)           
        for(j=i+1;j < n;j++)      
        {                         
            rhox[i][j] = 0.;      
            tmp       = (sumx12[i][i]-m*xbar[i]*xbar[i])*  
                        (sumx12[j][j]-m*xbar[j]*xbar[j]);             
            if (tmp > 0.)         
                rhox[i][j] = (sumx12[i][j] - m*xbar[i]*xbar[j]) / 
                             sqrt( tmp );      
            if (rhox[i][j] > 1.0)    rhox[i][j] = 1.0; 
            if (rhox[i][j] < -1.0)   rhox[i][j] = -1.0; 
        }                         
                                  
     /* free local arrays */      
     free_array(x,z,xbar,zbar,sumx,sumz,ssumx,ssumz,rhoz,
                sumx12,sumz12,ssumx12,ssumz12,sum1,sum2,sum3,sum4);
                                  
     return;                      
  }                               
                                  
                                  
  /* ...OTHERWISE, USE CONTROL-VARIATE ESTIMATE: 
  **    using micro/macro replications to estimate the correlations 
  **    of generated x's.  the total number of observations is m.  */ 
                                  
for(k=0;k < 20;k++)      /* start macro replications */
   {
     for(i=0;i < n; i++)    /* initialize the zero micro-sum */
     {
         sumz[i]   = 0.;
         sumx[i]   = 0.;
         for(j=i;j < n;j++)
         {
             sumz12[i][j] = 0.;
             sumx12[i][j] = 0.;
         }
     }

     for(l=0; l < batchsize; l++)    /* start micro replications */
     {

       multvar(iseed,idist,parm,n,chol,z,x);

       for(i=0;i < n;i++)
       {
           /* summation for z */
           sumz[i] = sumz[i] + z[i];
           for(j=i;j < n;j++)
               sumz12[i][j] = sumz12[i][j] + z[i] * z[j];

           /* summation for x */
           sumx[i] = sumx[i] + x[i];
           for(j=i;j < n;j++)
               sumx12[i][j] = sumx12[i][j] + x[i] * x[j];
       }

     }   /* End of micro replications */

     /* compute zbar and xbar */
     for(i=0;i < n;i++)
     {
        zbar[i] = sumz[i] / batchsize; 
        xbar[i] = sumx[i] / batchsize; 
     }

     /* compute rhoz and rhox estimates */
     for(i=0;i < n;i++)
        for(j=i+1;j < n;j++)
        {
            /* compute correlation estimates for z */
            tmp  =  (sumz12[i][i]-batchsize*zbar[i]*zbar[i])*
                    (sumz12[j][j]-batchsize*zbar[j]*zbar[j]);
            if (tmp > 0.)
                rhoz[i][j] = (sumz12[i][j]-batchsize*zbar[i]*zbar[j])
                             / sqrt( tmp );
            if (rhoz[i][j] > 1.0)    rhoz[i][j] = 1.0;
            if (rhoz[i][j] < -1.0)   rhoz[i][j] = -1.0;

            /* compute correlation estimates for x */
            tmp  =  (sumx12[i][i]-batchsize*xbar[i]*xbar[i])*
                    (sumx12[j][j]-batchsize*xbar[j]*xbar[j]);
            if (tmp > 0.)
                rhox[i][j] = (sumx12[i][j]-batchsize*xbar[i]*xbar[j])
                             / sqrt( tmp );
            if (rhox[i][j] > 1.0)    rhox[i][j] = 1.0;
            if (rhox[i][j] < -1.0)   rhox[i][j] = -1.0;
        }

     for(i=0;i < n;i++)
     {
        for(j=i;j < n;j++)
        {
            sum1[i][j] = sum1[i][j] + rhoz[i][j]; 
            sum2[i][j] = sum2[i][j] + rhox[i][j];
            sum3[i][j] = sum3[i][j] + rhoz[i][j]*rhox[i][j];
            sum4[i][j] = sum4[i][j] + rhoz[i][j]*rhoz[i][j];

            ssumz12[i][j] = ssumz12[i][j] + sumz12[i][j];
            ssumx12[i][j] = ssumx12[i][j] + sumx12[i][j];
        }
        ssumz[i] = ssumz[i] + sumz[i];
        ssumx[i] = ssumx[i] + sumx[i];
      }

   }  /* End of macro replications */

   for(i=0;i < n;i++)
   {
      zbar[i] = ssumz[i] / m;
      xbar[i] = ssumx[i] / m;
   }

   for(i=0;i < n;i++)
   {
      for(j=i+1;j < n;j++)
      {
          /* compute crude estimate of rhoz */
          tmp = (ssumx12[i][i]-m*xbar[i]*xbar[i])*
                (ssumx12[j][j]-m*xbar[j]*xbar[j]);
          if (tmp > 0.) 
              rhox[i][j] = (ssumx12[i][j]-m*xbar[i]*xbar[j])/ sqrt(tmp);

          /* compute crude correlation estimates for x */
          tmp = (ssumz12[i][i]-m*zbar[i]*zbar[i])*
                (ssumz12[j][j]-m*zbar[j]*zbar[j]);
          if (tmp > 0.)
              rhoz[i][j] = (ssumz12[i][j]-m*zbar[i]*zbar[j])/ sqrt(tmp);

          /* compute coefficients of control variates */
          k    = 20.;
          cov  = (sum3[i][j] - sum1[i][j]*sum2[i][j]/k) / (k-1);
          tmp  = (sum4[i][j] - sum1[i][j]*sum1[i][j]/k) / (k-1);
          beta = 1.;
          if (tmp != 0.)  beta = cov / tmp;

          /* compute the new correlation estimates for x */
          rhox[i][j] = rhox[i][j]-beta*(rhoz[i][j]-rho[i][j]);
          if (rhox[i][j] > 1.0)    rhox[i][j] = 1.0;
          if (rhox[i][j] < -1.0)   rhox[i][j] = -1.0;
      }
   }

   /* free local arrays */
   free_array(x,z,xbar,zbar,sumx,sumz,ssumx,ssumz,rhoz,
              sumx12,sumz12,ssumx12,ssumz12,sum1,sum2,sum3,sum4);
   return;
}


void corr_pair(long *iseed,int m,int *idist,double **parm,int n,
               double **rho,double **rhox,int *ifault)
{

   int     i,j,k,ndelete;
   double   **chol;

   chol   = (double**)malloc(n*sizeof(*chol));
   *chol  = (double *)malloc((n*n)*sizeof(**chol));
   for(i = 1;i < n;i++)
       *(chol+i) = *chol + i*n;

   /* rho is not positive definite, and then cannot be decomposed */
   /* solution: delete one row & one column from the bottom of rho
   **           until the submatrix is positive definite.         */

   ndelete = n;
   while( (*ifault) != 0)
   {
         ndelete--;
         choleskyf(ndelete,rho,chol,ifault);
   }

   /* Pairwisely, estimate correlations for random variables deleted */
   for (i = ndelete;i < n; i++)
        rowcorr(i,iseed,m,idist,parm,rho,rhox);

   corr_all(iseed,m,idist,parm,ndelete,rho,chol,rhox);

   free( *chol );
        free( chol );
}


void rowcorr(int n,long *iseed,int m,int *idist,double **parm,
             double **rho,double **rhox)
{

     int     i,j,k,l,ifault,batchsize;
     double   u,pr,d,cov,beta,tmp,
             *x,*z,*zind,*zbar,*xbar,**rhoz;
     double  *sumx,*sumz,**sumx12,**sumz12,**sum1,**sum2,**sum3,**sum4,
             *ssumx,*ssumz,**ssumx12,**ssumz12;

     /*  open local arrays  */
     open_array(n+1,&x,&z,&zbar,&xbar,&rhoz,&sumx,&sumz,
                &sumx12,&sumz12,&sum1,&sum2,&sum3,&sum4,&ssumx,
                &ssumz,&ssumx12,&ssumz12);
     zind   = (double*)malloc(n*sizeof(*zind));

     /*  initialize macro sums */
     for(i=0;i <= n;i++)
     {
         rhoz[i][n]     = 0.;
         rhox[i][n]     = 0.;
         ssumz[i]       = 0.;
         ssumx[i]       = 0.;
         sum1[i][n]     = 0.;
         sum2[i][n]     = 0.;
         sum3[i][n]     = 0.;
         sum4[i][n]     = 0.;
         ssumz12[i][i]  = 0.;
         ssumx12[i][i]  = 0.;
         ssumz12[i][n]  = 0.;
         ssumx12[i][n]  = 0.;
     }

     /* ...using micro-macro replications, generate m multivariates
     **    and compute correlation estimates. */
     batchsize = m / 20;      /* set macro replication number = 20 */

     for(k=0;k < 20;k++)      /* start macro replications */
     {
         for(i=0;i <= n; i++)    /* initialize micro-sum's */
         {
             sumz[i]      = 0.;
             sumx[i]      = 0.;
             sumz12[i][i] = 0.;
             sumx12[i][i] = 0.;
             sumz12[i][n] = 0.;
             sumx12[i][n] = 0.;
         }

         for(l=0; l < batchsize; l++)    /* start micro replications */
         {
           /************************************************************
           **  Generate X = (X_0,...,X_n),
           **    transformed from Z = (Z_0,...,Z_n)
           **    where corr(Z_i, Z_n) = rho[i][n], for i=0,...,n-1  
           ************************************************************/

           /* Step 1: Generate an independent standard-normal vector 
           **               Z = ( Z_0,...,Z_n )
           */
           for(i=0;i <= n;i++)
           {
               u = u16807d(iseed);
               zind[i] = ppnd(u,&ifault);
           }

           /* Step 2: Transform to correlated standard-normal vector
           **               Z = ( Z_0,...,Z_n )     so that
           **         corr( Z_i, Z_n ) = rho[i][n], for i=0,...,n-1
           */
           z[n] = zind[n];
           for(i=0;i < n;i++)
           {
               z[i] = rho[i][n] * zind[n] +
                      sqrt(1. - rho[i][n]*rho[i][n]) * zind[i];
           }

           /* Step 3: Transform to random vector 
           **               X = ( X_0,...,X_n ) 
           **         with specified marginal distributions F(x)  
           */
           for(i=0;i <= n;i++)
           {
               ndtr(z[i],&pr,&d);
               finv(idist[i],parm[i],pr,&x[i]);
           }
           /************************************************************
           **                     End of Generation                **
           ************************************************************/

           for(i=0;i <= n;i++)
           {
               sumz[i]      = sumz[i] + z[i];
               sumx[i]      = sumx[i] + x[i];
               sumz12[i][i] = sumz12[i][i] + z[i] * z[i];
               sumx12[i][i] = sumx12[i][i] + x[i] * x[i];
           }
           for(i=0;i < n;i++)
           {
               sumz12[i][n] = sumz12[i][n] + z[i] * z[n];
               sumx12[i][n] = sumx12[i][n] + x[i] * x[n];
           }
         }    /* End of micro replications */

         /* compute zbar and xbar */
         for(i=0;i <= n;i++)
         {
             zbar[i] = sumz[i] / batchsize;
             xbar[i] = sumx[i] / batchsize;
         }
         /* compute rhoz and rhox estimates */
         for(i=0;i < n;i++)
         {
             /* compute correlation estimates for z */
             tmp =  (sumz12[i][i]-batchsize*zbar[i]*zbar[i])*
                    (sumz12[n][n]-batchsize*zbar[n]*zbar[n]);
             if (tmp > 0.)
                 rhoz[i][n] =(sumz12[i][n]-batchsize*zbar[i]*zbar[n])
                                  / sqrt( tmp );
             if (rhoz[i][n] > 1.0)    rhoz[i][n] =  1.0;
             if (rhoz[i][n] < -1.0)   rhoz[i][n] = -1.0;

             /* compute correlation estimates for x */
             tmp  = (sumx12[i][i]-batchsize*xbar[i]*xbar[i])*
                    (sumx12[n][n]-batchsize*xbar[n]*xbar[n]);
             if (tmp > 0.)
                  rhox[i][n] = (sumx12[i][n]-batchsize*xbar[i]*xbar[n])
                                  / sqrt( tmp );
             if (rhox[i][n] > 1.0)    rhox[i][n] = 1.0;
             if (rhox[i][n] < -1.0)   rhox[i][n] = -1.0;
         }

         for(i=0;i <= n;i++)
         {
            ssumz[i] = ssumz[i] + sumz[i];
            ssumx[i] = ssumx[i] + sumx[i];
            ssumz12[i][i] = ssumz12[i][i] + sumz12[i][i];
            ssumx12[i][i] = ssumx12[i][i] + sumx12[i][i];
         }
         for(i=0;i < n;i++)
         {
            sum1[i][n] = sum1[i][n] + rhoz[i][n];
            sum2[i][n] = sum2[i][n] + rhox[i][n];
            sum3[i][n] = sum3[i][n] + rhoz[i][n]*rhox[i][n];
            sum4[i][n] = sum4[i][n] + rhoz[i][n]*rhoz[i][n];
            ssumz12[i][n] = ssumz12[i][n] + sumz12[i][n];
            ssumx12[i][n] = ssumx12[i][n] + sumx12[i][n];
         }

     }  /* End of macro replications */

     for(i=0;i <=n;i++)
     {
        zbar[i] = ssumz[i] / m;
        xbar[i] = ssumx[i] / m;
     }

     for(i=0;i < n;i++)
     {
          /* compute crude estimate of rhoz */
          tmp = (ssumz12[i][i]-m*zbar[i]*zbar[i])*
                (ssumz12[n][n]-m*zbar[n]*zbar[n]);
          if (tmp > 0.)
              rhoz[i][n] = (ssumz12[i][n]-m*zbar[i]*zbar[n])/sqrt(tmp);

          /* compute crude correlation estimates for x */
          tmp = (ssumx12[i][i]-m*xbar[i]*xbar[i])*
                (ssumx12[n][n]-m*xbar[n]*xbar[n]);
          if (tmp > 0.)
              rhox[i][n] = (ssumx12[i][n]-m*xbar[i]*xbar[n])/sqrt(tmp);

          /* compute coefficients of control variates */
          k    = 20.;
          cov  = (sum3[i][n] - sum1[i][n]*sum2[i][n]/k) / (k-1);
          tmp  = (sum4[i][n] - sum1[i][n]*sum1[i][n]/k) / (k-1);
          beta = 1.;
          if (tmp != 0.)  beta = cov / tmp;

          /* compute the new correlation estimates for x */
          rhox[i][n] = rhox[i][n] - beta * (rhoz[i][n] - rho[i][n]);
          if (rhox[i][n] > 1.0)    rhox[i][n] = 1.0;
          if (rhox[i][n] < -1.0)   rhox[i][n] = -1.0;
          rhox[n][i] = rhox[i][n];
      }
      rhox[n][n] = 1.;

      /* free local arrays */
      free_array(x,z,xbar,zbar,sumx,sumz,ssumx,ssumz,rhoz,
                 sumx12,sumz12,ssumx12,ssumz12,sum1,sum2,sum3,sum4);
      free( zind );
}


void  multvar(long *iseed,int *idist,double **parm,int n,double **chol,
              double *z,double *x)
/** purpose: generate a multivariate random vector x of dimension n,
             where each x_i has marginal distribution F_i (whose
             distribution parameters are stored in array parm),
             and the correlation matrix is rho (whose cholesky
             decompostion is chol, i.e., chol'*chol=rho).

    input:
            iseed: Initial seed
            parm:  Two-dimensional array.  Each parm[i] contains (np[i])
                   parameter values of distribution Fi, i=1,...,n
            n:     Dimension of x
            chol:  Cholesky decomposition of rho, s.t. chol'*chol=rho.

    output:
            iseed: Random-number seed
            z:     Generated standard-normal random vector
            x:     Generated multivariate vector                    **/

{
   int i;
   double  pr,d;

   /* ...step 1: generate a multivariate standard-normal vector with
                 correlation matrix rho                              */
   multnormal(iseed,n,chol,z);

   /* ...step 2 and 3: transform normal vectors to specified 
                       distributed vectors.  */
   for(i=0;i<n;i++)
   {
       ndtr(z[i],&pr,&d);                /* Step 2 */  
       finv(idist[i],parm[i],pr,&x[i]);  /* Step 3 */  
   }
   return;
}


void  multnormal(long *iseed,int n,double **chol,double *z)

/*  Purpose: generate a multivariate standard-normal random vector z
**           of size n, with correlation matrix rho.  (the cholesky
**           decomposition of rho is chol, i.e., chol'*chol=rho.)
**
**  Author:  Huifen Chen, Dyit-Yeh Tech., Taiwan.  11/19/1994.
**
**  Input:
**          iseed: initial seed
**          n:     dimension of Z
**          rho:   correlation matrix of Z
**
**  Output:
**          iseed: final seed
**          z: generated multivariate standard-normal             
*/

{
    double  u,*zind;
    int i,j,ifault;

    zind = (double *)malloc(n*sizeof(*zind));

    /* ...generate a vector of n independent standard-normal variates.  */
    for(i=0;i<n;i++)
    {
        u = u16807d(iseed);
        zind[i] = ppnd(u,&ifault);
    }

    /*  ....generate a multivariate stardard-normal vector with
            correlation rho.  let z <-- chol_transpose * z.     */
    for(i = 0;i < n;i++)
    {
        z[i] = 0.0;
        for(j = 0;j < i+1;j++)
              z[i] = z[i] + chol[j][i] * zind[j];
    }

    free( zind );
    return;
}




double u16807d(long *iseed)
{
   double u;
   do
   {
        *iseed =(long) fmod((double)(*iseed)*16807.0, 2147483647.0);
        u = (*iseed) / 2147483648.0;
   }
   while((u >= 1.0) || ((1.0-u) >= 1.0) );
   return(u);
}


/***********************************************************************
** FUNCTION: ppnd(p, ifault)                                          **
**                                                                    **
** Purpose:  Compute the inversed distribution function of a          **
**           standard normal distribution that is set to the          **
**           value of z satisfying P(Z<z) = p. therefore,             **
**           0 < p < 1.                                               **
**                                                                    **
** Input:                                                             **
**           p:      the percentage point to evaluate                 **
**                                                                    **
** Output:                                                            **
**           ppnd:   the inverse cdf                                  **
**           ifault: error indicator. ifault=0 means that all is well.**
**                                                                    **
** Reference:                                                         **
**           Beasley, J.D. and Springer, S.G. (1977).                 **
**           Algorithm AS 111.  The Percentage Points of the Normal   **
**           Distribution.  Applied Statistics v 26, 118-121.         **
**                                                                    **
***********************************************************************/

double ppnd(double p,int *ifault )
{
   double ppnd1,
         q,
         zero   = 0.,
         split  = 0.42,
         half   = 0.5,
         one    = 1.,
         r,
          a0    =   2.50662823884,
          a1    = -18.61500062529,
          a2    =  41.39119773534,
          a3    = -25.44106049637,
          b1    =  -8.47351093090,
          b2    =  23.08336743743,
          b3    = -21.06224101826,
          b4    =   3.13082909833,
          c0    =  -2.78718931138,
          c1    =  -2.29796479134,
          c2    =   4.85014127135,
          c3    =   2.32121276858,
          d1    =   3.54388924762,
          d2    =   1.63706781897;
    double tmp;

    *ifault = 0;
    q = p - half;
    tmp = fabs((double)q);
    if( tmp > split )
    {
        r = p;
        if ( q > zero ) r = one - p;
        if ( r <= zero )
        {
            *ifault = 1;
            ppnd1 = zero;
            return(ppnd1);
        }
        else
        {
            r = sqrt(-log(r));
            ppnd1 = ((( c3 * r + c2 ) * r + c1 ) * r + c0 );
            ppnd1 = ppnd1 / (( d2 * r + d1 ) * r + one );
            if ( q < 0 ) ppnd1 = -ppnd1;
            return(ppnd1);
        }
    }
    else
    {
        r = q * q;
        ppnd1 = q * ((( a3 * r + a2 ) * r + a1 ) * r + a0 );
        ppnd1 = ppnd1 / (((( b4 * r + b3 ) * r + b2 ) * r + b1 )
                * r + one );
        return( ppnd1 );
    }
}


void ndtr(double x,double *p,double *d)

/*    compute the cumulative distribution function P(X <= x)
*       of a standard normal random variable X.
*     input 
*       x: normal value 
*     output
*       p: P(X <= x) 
*       d: standard normal density at x 
*     reference 
*       c. hastings, approximations for digital computers. 
*       princeton university press, princeton, new jersey, usa, 1955.
*       this code is a minor modification, by bruce schmeiser, 
*       of the code in the ibm scientific subroutine package, 
*       1967, page 78. 
*/

{
   double t,ax;

   ax = fabs((double) x);
   t = 1.0 / (ax * .2316419 + 1.0);
   *d = exp( -(x * x * .5) ) * .3989423;
   *p = 1.0 - *d * t * ((((t * 1.330274 - 1.821256) * t
            + 1.781478) * t - .3565638) * t + .3193815);
   if (x < 0.0) *p = 1 - *p;
   return;
}



void open_array(int n,double **x,double **z,double **zbar,double **xbar,
                double ***rhoz,double **sumx,double **sumz,double ***sumx12,
                double ***sumz12,double ***sum1,double ***sum2,double ***sum3,
                double ***sum4,double **ssumx,double **ssumz,double ***ssumx12,
                double ***ssumz12)

{

   int i;

   /*  array declaration  */
   *z      = (double*)malloc(n*sizeof(**z));
   *x      = (double*)malloc(n*sizeof(**x));
   *zbar   = (double  *)malloc(n*sizeof(**zbar));
   *xbar   = (double  *)malloc(n*sizeof(**xbar));
   *rhoz= (double **)malloc(n*sizeof(**rhoz));
   **rhoz= (double *)malloc((n*n)*sizeof(***rhoz));
   for(i=1;i < n;i++)
       *(*rhoz+i)= **rhoz + i*n;
   *sumz   = (double *)malloc(n*sizeof(**sumz));
   *sumz12 = (double **)malloc(n*sizeof(**sumz12));
   **sumz12= (double *)malloc((n*n)*sizeof(***sumz12));
   for(i=1;i < n;i++)
       *(*sumz12+i)= **sumz12 + i*n;
   *sumx    = (double *)malloc(n*sizeof(**sumx));
   *sumx12  = (double **)malloc(n*sizeof(**sumx12));
   **sumx12 = (double *)malloc((n*n)*sizeof(***sumx12));
   for(i=1;i < n;i++)
       *(*sumx12+i)= **sumx12 + i*n;
   *ssumz   = (double *)malloc(n*sizeof(**ssumz));
   *ssumz12 = (double **)malloc(n*sizeof(**ssumz12));
   **ssumz12= (double *)malloc((n*n)*sizeof(***ssumz12));
   for(i=1;i < n;i++)
       *(*ssumz12+i)= **ssumz12 + i*n;
   *ssumx    = (double *)malloc(n*sizeof(**ssumx));
   *ssumx12  = (double **)malloc(n*sizeof(**ssumx12));
   **ssumx12 = (double *)malloc((n*n)*sizeof(***ssumx12));
   for(i=1;i < n;i++)
       *(*ssumx12+i)= **ssumx12 + i*n;

   *sum1  = (double **)malloc(n*sizeof(**sum1));
   *sum2  = (double **)malloc(n*sizeof(**sum2));
   *sum3  = (double **)malloc(n*sizeof(**sum3));
   *sum4  = (double **)malloc(n*sizeof(**sum4));
   **sum1 = (double *)malloc((n*n)*sizeof(***sum1));
   **sum2 = (double *)malloc((n*n)*sizeof(***sum2));
   **sum3 = (double *)malloc((n*n)*sizeof(***sum3));
   **sum4 = (double *)malloc((n*n)*sizeof(***sum4));
   for(i=1;i < n;i++)
   {
       *(*sum1+i)= **sum1 + i*n;
       *(*sum2+i)= **sum2 + i*n;
       *(*sum3+i)= **sum3 + i*n;
       *(*sum4+i)= **sum4 + i*n;
   }

}


void free_array(double *x,double *z,double *xbar,double *zbar,double *sumx,
                double *sumz,double *ssumx,double *ssumz,double **rhoz,
                double **sumx12,double **sumz12,double **ssumx12,
                double **ssumz12,double **sum1,double **sum2,double **sum3,
                double **sum4)
{
 
     free( x );
     free( z );
     free( xbar );
     free( zbar );
     free( sumx );
     free( sumz );
     free( ssumx );
     free( ssumz );
     free( *rhoz );
         free( rhoz );
     free( *sumx12 );
         free( sumx12 );
     free( *sumz12 );
         free( sumz12 );
     free( *ssumx12 );
         free( ssumx12 );
     free( *ssumz12 );
       free( ssumz12 );
     free( *sum1 );
         free( sum1 );
     free( *sum2 );
         free( sum2 );
     free( *sum3 );
         free( sum3 );
     free( *sum4 );
         free( sum4 );
}

void free_2darray(int m, int n, double **a)

/* Purpose: free a two-dimensional array a.  the size is m x n. */
{
    int i;

    for( i=0; i< m; i++ )
         free( a[i] );
    free( a );
}

